#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include "dw.h"

int _dw_errno = DW_ERR_NONE;
int _dw_sym_base  = 0;
int _dw_sym_count = 0;
int _dw_sym_size  = 0;
dw_syminfo_t** _dw_sym_table = NULL;

const char* _dw_errstr[] =
{
  "no error",
  "dlopen() failed",
  "out of memory",
  "invalid library handle",
  "bad pointer",
  "library has already a symbol table"
};

int _dw_symtab_add (dw_syminfo_t* sym);


/*
 * Public interface
 */

dw_library_t* dw_open (const char* name, const char* dllpath)
{
  void* dll;
  dw_library_t* lib;

  dll = dlopen (dllpath, RTLD_NOW);
  if (dll == NULL)
  {
    _dw_errno = DW_ERR_DLOPEN;
    return NULL;
  }
  lib = (dw_library_t*) malloc (sizeof(dw_library_t));
  if (!lib)
  {
    dlclose (dll);
    _dw_errno = DW_ERR_OOM;
    return NULL;
  }
  strncpy (lib->name, name, FILENAME_MAX-1); lib->name[FILENAME_MAX-1] = 0;
  strncpy (lib->dll_path, dllpath, FILENAME_MAX-1); lib->dll_path[FILENAME_MAX-1] = 0;
  lib->dll_handle = dll;
  lib->sym_base = 0;
  lib->sym_nb = 0;
  return lib;
}

int dw_add (dw_library_t* lib, dw_syminfo_t* symbols, int count)
{
  int i;

  if (lib == NULL || lib->dll_handle == NULL)
  {
    _dw_errno = DW_ERR_BADLIB;
    return _dw_errno;
  }
  if (symbols == NULL)
  {
    _dw_errno = DW_ERR_BADPTR;
    return _dw_errno;
  }
  if (lib->sym_nb > 0)
  {
    _dw_errno = DW_ERR_HASSYM;
    return _dw_errno;
  }

  i = 0;
  lib->sym_base = _dw_sym_count;
  while ((count > 0) ? (i < count) : (symbols[i].name != NULL))
  {
    void* func;

    func = dlsym (lib->dll_handle, symbols[i].name);
    if (func == NULL)
    {
      *symbols[i].func = NULL;
    }
    else
    {
      *symbols[i].func = func;
      if (_dw_symtab_add (&(symbols[i]))) return _dw_errno;
    }
    i++;
  }
  lib->sym_nb = i;
  return DW_ERR_NONE;
}

int dw_dump (dw_library_t* lib)
{
  if (lib == NULL || lib->dll_handle == NULL)
  {
    _dw_errno = DW_ERR_BADLIB;
    return _dw_errno;
  }
  printf ("%s: wrapping %s (%d symbols)\n", lib->name, lib->dll_path, lib->sym_nb);
  return DW_ERR_NONE;
}

int dw_close (dw_library_t* lib)
{
  if (lib == NULL)
  {
    _dw_errno = DW_ERR_BADLIB;
    return _dw_errno;
  }
  /* FIXME: release a lot of stuff ? warn listeners ? */
  if (lib->dll_handle) dlclose (lib->dll_handle);
  free (lib);
  return 0;
}

const char* dw_error ()
{
  return (_dw_errno == DW_ERR_DLOPEN) ? dlerror() : _dw_errstr[_dw_errno];
}

void dw_do (dw_syminfo_t* syminfo, ...)
{
  printf ("do  : %s\n", syminfo->name);
}

void dw_do_r (dw_syminfo_t* syminfo, dw_ret_t ret, ...)
{
  printf ("do_r: %s: %x\n", syminfo->name, (int)ret);
}

void dw_na (dw_syminfo_t* syminfo, ...)
{
  printf ("na  : %s\n", syminfo->name);
}


/*
 * Internals
 */

#define DW_CHUNK_SIZE 128

int _dw_symtab_add (dw_syminfo_t* sym)
{
  _dw_sym_count++;
  if (_dw_sym_table == NULL || _dw_sym_count > _dw_sym_size)
  {
    _dw_sym_size += DW_CHUNK_SIZE;
    _dw_sym_table = (dw_syminfo_t**) realloc (_dw_sym_table, sizeof(dw_syminfo_t*) * _dw_sym_size);
    if (_dw_sym_table == NULL)
    {
      _dw_errno = DW_ERR_OOM;
      return _dw_errno;
    }
  }
  _dw_sym_table[_dw_sym_count++] = sym;
  return DW_ERR_NONE;
}
