#include <stdlib.h>
#include "dw.h"
#include "gl_wrapper.h"
#include "glx_wrapper.h"
#include "glu_wrapper.h"

#define GLTUNE_VERSION "0.0.1"

dw_library_t* gl_lib;
dw_library_t* glx_lib;
dw_library_t* glu_lib;


dw_library_t*  _gltune_wrap (char* varname, char* preset, char* name, dw_syminfo_t* symtab)
{
  dw_library_t* lib;
  char* env;
  char* path;

  env = getenv (varname);
  path = (env) ? env : preset;

  if ((lib = dw_open (name, path)) == NULL ||
      (dw_add (lib, symtab, 0)))
  {
    printf ("error: %s\n", dw_error());
  }
  dw_dump (lib);
  return lib;
}

void _init()
{
  const char* unset = "(unset)";
  char* ld_preload;
  char* ld_library_path;
  int i;

  printf ("\n*\n*  gltune " GLTUNE_VERSION "\n*\n\n");

  ld_preload = getenv ("LD_PRELOAD");
  ld_library_path = getenv ("LD_LIBRARY_PATH");
  printf ("LD_PRELOAD     : %s\n", ld_preload ? ld_preload : unset);
  printf ("LD_LIBRARY_PATH: %s\n\n", ld_library_path ? ld_library_path : unset);

  gl_lib  = _gltune_wrap ("GLTUNE_GL_PATH", "/usr/lib/libGL.so", "GL", _symtab_gl);
  glx_lib = _gltune_wrap ("GLTUNE_GL_PATH", "/usr/lib/libGL.so", "GLX", _symtab_glx);
  glu_lib = _gltune_wrap ("GLTUNE_GLU_PATH", "/usr/lib/libGLU.so", "GLU", _symtab_glu);
  printf("\n");

  if (getenv ("GLTUNE_TRACE"))
    for (i = 0; i < glx_lib->sym_nb; i++) _symtab_glx[i].flags |= DW_FLAG_TRACE;
}

void _fini()
{
  dw_close (glu_lib);
  dw_close (glx_lib);
  dw_close (gl_lib);
}
